////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Audio/SoundFileOgg.hpp>
#include <iostream>


namespace sf
{
namespace priv
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
SoundFileOgg::SoundFileOgg() :
myFile(NULL)
{

}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
SoundFileOgg::~SoundFileOgg()
{
    if (myFile)
    {
        ov_clear(&myStream);
        myFile = NULL;
    }
}


////////////////////////////////////////////////////////////
/// Open the sound file for reading
////////////////////////////////////////////////////////////
bool SoundFileOgg::OpenRead(const std::string& Filename, std::size_t& NbSamples, unsigned int& ChannelsCount, unsigned int& SampleRate)
{
    // Close the file if already opened
    if (myFile)
    {
        ov_clear(&myStream);
        myFile = NULL;
    }

    // Open the file
    myFile = fopen(Filename.c_str(), "rb");
    if (!myFile)
    {
        std::cerr << "Failed to read sound file \"" << Filename << "\" (cannot open the file)" << std::endl;
        return false;
    }

    // Bind the file to the ogg-vorbis stream
    int Error = ov_open(myFile, &myStream, NULL, 0);
    if (Error < 0)
    {
        std::cerr << "Failed to read sound file \"" << Filename << "\" (not a valid ogg-vorbis file)" << std::endl;
        fclose(myFile);
        myFile = NULL;
        return false;
    }

    // Set the sound parameters
    vorbis_info* Infos = ov_info(&myStream, -1);
    ChannelsCount = Infos->channels;
    SampleRate    = Infos->rate;
    NbSamples     = static_cast<std::size_t>(ov_pcm_total(&myStream, -1) * ChannelsCount);

    // Make sure that the ogg stream has a fixed bitrate (VBR is not supported yet...)
    /*if ((Infos->bitrate_lower != Infos->bitrate_nominal) || (Infos->bitrate_lower != Infos->bitrate_upper))
    {
        std::cerr << "Failed to read sound file \"" << Filename << "\" (variable bitrate is not supported)" << std::endl;
        return false;
    }*/

    return true;
}


////////////////////////////////////////////////////////////
/// Read samples from the loaded sound
////////////////////////////////////////////////////////////
std::size_t SoundFileOgg::Read(Int16* Data, std::size_t NbSamples)
{
    // ov_read will only read one chunk, even if it is smaller than the requested size,
    // so we loop until all bytes have been read
    long TotalSize = static_cast<long>(NbSamples * sizeof(Int16));
    long TotalRead = 0;
    while (TotalRead < TotalSize)
    {
        // Decode sound data
        char* Ptr = reinterpret_cast<char*>(Data) + TotalRead;
        int   Len = TotalSize - TotalRead;
        long Read = ov_read(&myStream, Ptr, Len, 0, 2, 1, NULL);

        if (Read > 0)
        {
            // Ok, Read bytes have been read
            TotalRead += Read;
        }
        else if (Read == 0)
        {
            // End of file reached
            break;
        }
        else
        {
            // Error...
            std::cerr << "Failed to read from ogg file" << std::endl;
            break;
        }
    }

    return static_cast<std::size_t>(TotalRead / sizeof(Int16));
}

} // namespace priv

} // namespace sf
