////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/Color.hpp>
#include <algorithm>


namespace sf
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
const Color Color::Black(0, 0, 0);
const Color Color::White(255, 255, 255);
const Color Color::Red(255, 0, 0);
const Color Color::Green(0, 255, 0);
const Color Color::Blue(0, 0, 255);
const Color Color::Yellow(255, 255, 0);
const Color Color::Magenta(255, 0, 255);
const Color Color::Cyan(0, 255, 255);


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
Color::Color() :
r(0),
g(0),
b(0),
a(255)
{

}


////////////////////////////////////////////////////////////
/// Construct the color from a RGBA 32 bits integer
////////////////////////////////////////////////////////////
Color::Color(Uint32 ColorRGBA) :
r(static_cast<Uint8>((ColorRGBA & 0x000000FF))),
g(static_cast<Uint8>((ColorRGBA & 0x0000FF00) >> 8)),
b(static_cast<Uint8>((ColorRGBA & 0x00FF0000) >> 16)),
a(static_cast<Uint8>((ColorRGBA & 0xFF000000) >> 24))
{

}


////////////////////////////////////////////////////////////
/// Construct the color from its 4 RGBA components
////////////////////////////////////////////////////////////
Color::Color(Uint8 R, Uint8 G, Uint8 B, Uint8 A) :
r(R),
g(G),
b(B),
a(A)
{

}


////////////////////////////////////////////////////////////
/// Get the RGBA representation of the color
////////////////////////////////////////////////////////////
Uint32 Color::ToRGBA() const
{
    return (a << 24) | (b << 16) | (g << 8) | r;
}


////////////////////////////////////////////////////////////
/// Compare two colors (for equality)
////////////////////////////////////////////////////////////
bool Color::operator ==(const Color& Other) const
{
    return (r == Other.r) && (g == Other.g) && (b == Other.b) && (a == Other.a);
}


////////////////////////////////////////////////////////////
/// Compare two colors (for difference)
////////////////////////////////////////////////////////////
bool Color::operator !=(const Color& Other) const
{
    return (r != Other.r) || (g != Other.g) || (b != Other.b) || (a != Other.a);
}


////////////////////////////////////////////////////////////
/// Operator + overload to add two colors
////////////////////////////////////////////////////////////
Color operator +(const Color& Color1, const Color& Color2)
{
    int R = std::min(Color1.r + Color2.r, 255);
    int G = std::min(Color1.g + Color2.g, 255);
    int B = std::min(Color1.b + Color2.b, 255);
    int A = std::min(Color1.a + Color2.a, 255);

    return Color(static_cast<Uint8>(R), static_cast<Uint8>(G), static_cast<Uint8>(B), static_cast<Uint8>(A));
}


////////////////////////////////////////////////////////////
/// Operator * overload to modulate two colors
////////////////////////////////////////////////////////////
Color operator *(const Color& Color1, const Color& Color2)
{
    int R = Color1.r * Color2.r / 255;
    int G = Color1.g * Color2.g / 255;
    int B = Color1.b * Color2.b / 255;
    int A = Color1.a * Color2.a / 255;

    return Color(static_cast<Uint8>(R), static_cast<Uint8>(G), static_cast<Uint8>(B), static_cast<Uint8>(A));
}

} // namespace sf
