////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/Drawable.hpp>
#include <SFML/Graphics/OpenGL.hpp>


namespace sf
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
Drawable::Drawable(float Left, float Top, float ScaleX, float ScaleY, float Rotation, const Color& Col) :
myLeft    (Left),
myTop     (Top),
myScaleX  (ScaleX),
myScaleY  (ScaleY),
myRotation(Rotation),
myCenterX (0.f),
myCenterY (0.f),
myColor   (Col)
{

}


////////////////////////////////////////////////////////////
/// Virtual destructor
////////////////////////////////////////////////////////////
Drawable::~Drawable()
{
    // Nothing to do
}


////////////////////////////////////////////////////////////
/// Set the left position of the object
////////////////////////////////////////////////////////////
void Drawable::SetLeft(float Left)
{
    myLeft = Left;
}


////////////////////////////////////////////////////////////
/// Set the top position of the object
////////////////////////////////////////////////////////////
void Drawable::SetTop(float Top)
{
    myTop = Top;
}


////////////////////////////////////////////////////////////
/// Set the scale of the object
////////////////////////////////////////////////////////////
void Drawable::SetScale(float ScaleX, float ScaleY)
{
    if (ScaleX > 0.f) myScaleX = ScaleX;
    if (ScaleY > 0.f) myScaleY = ScaleY;
}


////////////////////////////////////////////////////////////
/// Set the orientation of the object
////////////////////////////////////////////////////////////
void Drawable::SetRotation(float Rotation)
{
    myRotation = Rotation;
}


////////////////////////////////////////////////////////////
/// Set the center of rotation, in coordinates relative to the
/// object
////////////////////////////////////////////////////////////
void Drawable::SetRotationCenter(float X, float Y)
{
    myCenterX = X;
    myCenterY = Y;
}


////////////////////////////////////////////////////////////
/// Set the color of the object
////////////////////////////////////////////////////////////
void Drawable::SetColor(const Color& Col)
{
    myColor = Col;
}


////////////////////////////////////////////////////////////
/// Get the left position of the object
////////////////////////////////////////////////////////////
float Drawable::GetLeft() const
{
    return myLeft;
}


////////////////////////////////////////////////////////////
/// Get the top position of the object
////////////////////////////////////////////////////////////
float Drawable::GetTop() const
{
    return myTop;
}


////////////////////////////////////////////////////////////
/// Get the horizontal scale of the object
////////////////////////////////////////////////////////////
float Drawable::GetScaleX() const
{
    return myScaleX;
}


////////////////////////////////////////////////////////////
/// Get the vertical scale of the object
////////////////////////////////////////////////////////////
float Drawable::GetScaleY() const
{
    return myScaleY;
}


////////////////////////////////////////////////////////////
/// Get the orientation of the object
////////////////////////////////////////////////////////////
float Drawable::GetRotation() const
{
    return myRotation;
}


////////////////////////////////////////////////////////////
/// Get the color of the object
////////////////////////////////////////////////////////////
const Color& Drawable::GetColor() const
{
    return myColor;
}


////////////////////////////////////////////////////////////
/// Move the object
////////////////////////////////////////////////////////////
void Drawable::Move(float OffsetX, float OffsetY)
{
    myLeft += OffsetX;
    myTop  += OffsetY;
}


////////////////////////////////////////////////////////////
/// Scale the object
////////////////////////////////////////////////////////////
void Drawable::Scale(float FactorX, float FactorY)
{
    if (FactorX > 0.f) myScaleX *= FactorX;
    if (FactorY > 0.f) myScaleY *= FactorY;
}


////////////////////////////////////////////////////////////
/// Rotate the object
////////////////////////////////////////////////////////////
void Drawable::Rotate(float Angle)
{
    myRotation += Angle;
}


////////////////////////////////////////////////////////////
/// Draw the object into the specified window
////////////////////////////////////////////////////////////
void Drawable::Draw(RenderWindow& Window)
{
    // Save the current modelview matrix
    GLCheck(glMatrixMode(GL_MODELVIEW));
    GLCheck(glPushMatrix());

    // Setup alpha-blending
    GLCheck(glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA));

    // Set color
    GLCheck(glColor4ub(myColor.r, myColor.g, myColor.b, myColor.a));

    // Set transformations
    GLCheck(glTranslatef(myLeft + myCenterX, myTop + myCenterY, 0));
    GLCheck(glRotatef(-myRotation, 0, 0, 1));
    GLCheck(glTranslatef(-myCenterX, -myCenterY, 0));
    GLCheck(glScalef(myScaleX, myScaleY, 1));

    // Let the derived class render the object geometry
    Render(Window);

    // Restore the previous modelview matrix
    GLCheck(glPopMatrix());
}

} // namespace sf
