////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/ImageLoader.hpp>
#include <iostream>
#undef _UNICODE
#include <IL/il.h>


namespace sf
{
namespace priv
{
////////////////////////////////////////////////////////////
/// Get the unique instance of the class
////////////////////////////////////////////////////////////
ImageLoader& ImageLoader::GetInstance()
{
    static ImageLoader Instance;

    return Instance;
}


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
ImageLoader::ImageLoader()
{
    // Initialize DevIL
    ilInit();
    ILenum ErrorCode = ilGetError();
    if (ErrorCode != IL_NO_ERROR)
    {
        std::cerr << "Failed to initialize DevIL library (error code : " << ErrorCode << ")" << std::endl;
        return;
    }

    // Check DevIL version
    ILint ILVersion = ilGetInteger(IL_VERSION_NUM);
    if (ILVersion < IL_VERSION)
    {
        std::cerr << "DevIL library has incorrect version (is " << ILVersion << ", should be " << IL_VERSION << ")" << std::endl;
        return;
    }

    // Allow files to be overwritten by save operations if they already exist
    ilEnable(IL_FILE_OVERWRITE);

    // Force image origin to upper left corner
    ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
    ilEnable(IL_ORIGIN_SET);

    // Force load format to 32 bits RGBA
    ilSetInteger(IL_FORMAT_MODE, IL_RGBA);
    ilEnable(IL_FORMAT_SET);
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
ImageLoader::~ImageLoader()
{
    // Shutdown DevIL library
    ilShutDown();
}


////////////////////////////////////////////////////////////
/// Load pixels from an image file
////////////////////////////////////////////////////////////
bool ImageLoader::LoadImageFromFile(const std::string& Filename, std::vector<Uint32>& Pixels, unsigned int& Width, unsigned int& Height)
{
    // First generate a new DevIL image
    ILuint Image;
    ilGenImages(1, &Image);
    ilBindImage(Image);

    // Load from file
    if (ilLoadImage(const_cast<char*>(Filename.c_str())) == false)
    {
        std::cerr << "Failed to load image from file \"" << Filename << "\"" << std::endl;
        ilDeleteImages(1, &Image);
        return false;
    }

    // Get image dimensions
    Width  = ilGetInteger(IL_IMAGE_WIDTH);
    Height = ilGetInteger(IL_IMAGE_HEIGHT);

    // Copy pixels to pixel buffer
    const Uint32* PixelsPtr = reinterpret_cast<const Uint32*>(ilGetData());
    Pixels.assign(PixelsPtr, PixelsPtr + Width * Height);

    // Destroy DevIL image, we don't need it anymore
    ilDeleteImages(1, &Image);

    return true;
}


////////////////////////////////////////////////////////////
/// Save pixels to an image file
////////////////////////////////////////////////////////////
bool ImageLoader::SaveImageToFile(const std::string& Filename, const std::vector<Uint32>& Pixels, unsigned int Width, unsigned int Height)
{
    // Generate a new DevIL image
    ILuint Image;
    ilGenImages(1, &Image);
    ilBindImage(Image);

    // Set its size and format to match those of the image to save
    if (!ilTexImage(Width, Height, 1, 4, IL_RGBA, IL_UNSIGNED_BYTE, const_cast<Uint32*>(&Pixels[0])))
    {
        std::cerr << "Failed to save image to file \"" << Filename << "\"" << std::endl;
        ilDeleteImages(1, &Image);
        return false;
    }

    // Save image
    if (!ilSaveImage(const_cast<char*>(Filename.c_str())))
    {
        std::cerr << "Failed to save image to file \"" << Filename << "\"" << std::endl;
        ilDeleteImages(1, &Image);
        return false;
    }

    // Destroy DevIL image, we don't need it anymore
    ilDeleteImages(1, &Image);

    return true;
}

} // namespace priv

} // namespace sf
