////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/RenderWindow.hpp>
#include <SFML/Graphics/Drawable.hpp>
#include <SFML/Graphics/GraphicsDevice.hpp>
#include <SFML/Graphics/Image.hpp>
#include <SFML/Graphics/OpenGL.hpp>
#include <iostream>


namespace sf
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
RenderWindow::RenderWindow() :
myBackgroundColor(Color(0, 0, 0, 255)),
myOpenGLMode     (false)
{

}


////////////////////////////////////////////////////////////
/// Construct the window
////////////////////////////////////////////////////////////
RenderWindow::RenderWindow(VideoMode Mode, const std::string& Title, RenderWindow::Style WindowStyle, int AntialiasingLevel) :
myBackgroundColor(Color(0, 0, 0, 255)),
myOpenGLMode     (false)
{
    Create(Mode, Title, WindowStyle, AntialiasingLevel);
}


////////////////////////////////////////////////////////////
/// Construct the window from an existing control
////////////////////////////////////////////////////////////
RenderWindow::RenderWindow(WindowHandle Handle, int AntialiasingLevel) :
myBackgroundColor(Color(0, 0, 0, 255)),
myOpenGLMode     (false)
{
    Create(Handle, AntialiasingLevel);
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
RenderWindow::~RenderWindow()
{
    // Nothing to do...
}


////////////////////////////////////////////////////////////
/// Create the window
////////////////////////////////////////////////////////////
void RenderWindow::Create(VideoMode Mode, const std::string& Title, RenderWindow::Style WindowStyle, int AntialiasingLevel)
{
    // Let the base class do the job
    Window::Create(Mode, Title, WindowStyle, AntialiasingLevel);

    // Initialize the rendering states
    Initialize();
}


////////////////////////////////////////////////////////////
/// Create the window from an existing control
////////////////////////////////////////////////////////////
void RenderWindow::Create(WindowHandle Handle, int AntialiasingLevel)
{
    // Let the base class do the job
    Window::Create(Handle, AntialiasingLevel);

    // Initialize the rendering states
    Initialize();
}


////////////////////////////////////////////////////////////
/// Display window content on screen
////////////////////////////////////////////////////////////
void RenderWindow::Display()
{
    // Discard display if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot display a window while in OpenGL mode ; call EndOpenGL() before calling Display()" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (Window::SetCurrent())
    {
        // Display backbuffer on screen
        Window::Display();

        // Find which buffers we must clear
        GLbitfield                ClearBits  = GL_COLOR_BUFFER_BIT;
        if (GetDepthBits()   > 0) ClearBits |= GL_DEPTH_BUFFER_BIT;
        if (GetStencilBits() > 0) ClearBits |= GL_STENCIL_BUFFER_BIT;
        
        // Clear the color/depth/stencil buffers for next frame
        GLCheck(glClearColor(myBackgroundColor.r / 255.f,
                             myBackgroundColor.g / 255.f,
                             myBackgroundColor.b / 255.f,
                             myBackgroundColor.a / 255.f));
        GLCheck(glClear(ClearBits));

        // Reset transformations
        GLCheck(glMatrixMode(GL_MODELVIEW));
        GLCheck(glLoadIdentity());
    }
}


////////////////////////////////////////////////////////////
/// Draw something on the window
////////////////////////////////////////////////////////////
void RenderWindow::Draw(Drawable& Object)
{
    // Discard draw if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot use SFML rendering while in OpenGL mode ; call EndOpenGL() before calling Draw()" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (Window::SetCurrent())
    {
        // Let the object draw itself
        Object.Draw(*this);
    }
}


////////////////////////////////////////////////////////////
/// Save the content of the window to an image
////////////////////////////////////////////////////////////
Image RenderWindow::Capture() const
{
    // Get the window dimensions
    const unsigned int Width  = GetWidth();
    const unsigned int Height = GetHeight();

    // Set our window as the current target for rendering
    if (Window::SetCurrent())
    {
        // Get pixels from the backbuffer
        std::vector<Uint32> Pixels(Width * Height);
        glReadPixels(0, 0, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, &Pixels[0]);

        // Create an image from it and return it
        return Image(Width, Height, &Pixels[0]);
    }
    else
    {
        return Image(Width, Height, Color::White);
    }
}


////////////////////////////////////////////////////////////
/// Change the background color of the window
////////////////////////////////////////////////////////////
void RenderWindow::SetBackgroundColor(const Color& Col)
{
    myBackgroundColor = Col;
}


////////////////////////////////////////////////////////////
/// Get the event on top of events stack, if any
////////////////////////////////////////////////////////////
bool RenderWindow::GetEvent(Event& EventReceived)
{
    // Forward to base class
    if (Window::GetEvent(EventReceived))
    {
        // Intercept resize events, because we have to adjust the viewport to new size
        if (EventReceived.Type == Event::Resize)
            GLCheck(glViewport(0, 0, EventReceived.Size.Width, EventReceived.Size.Height));

        return true;
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Change the current active view
////////////////////////////////////////////////////////////
void RenderWindow::SetView(const View* NewView)
{
    // Discard display if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot change the view while in OpenGL mode ; call EndOpenGL() before setting a new SFML view" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (Window::SetCurrent())
    {
        // If a null pointer is passed, switch to default view
        if (NewView == NULL)
            NewView = &myDefaultView;

        // Compute view rectangle coordinates
        float X = NewView->Left + NewView->Width  / 2;
        float Y = NewView->Top  + NewView->Height / 2;
        float HalfWidth  = NewView->Width  / (2 * NewView->Zoom);
        float HalfHeight = NewView->Height / (2 * NewView->Zoom);

        // Store view rectangle for optimization purpose
        myCurrentRect = FloatRect(X - HalfWidth, Y - HalfHeight, X + HalfWidth, Y + HalfHeight);

        // Update the projection matrix according to the new view
        GLCheck(glMatrixMode(GL_PROJECTION));
        GLCheck(glLoadIdentity());
        GLCheck(glOrtho(myCurrentRect.Left, myCurrentRect.Right, myCurrentRect.Bottom, myCurrentRect.Top, -1, 1));
    }
}


////////////////////////////////////////////////////////////
/// Get current view rectangle
////////////////////////////////////////////////////////////
const FloatRect& RenderWindow::GetViewRect() const
{
    return myCurrentRect;
}


////////////////////////////////////////////////////////////
/// Start custom OpenGL rendering
////////////////////////////////////////////////////////////
void RenderWindow::BeginOpenGL()
{
    if (!myOpenGLMode && Window::SetCurrent())
    {
        // Save the current matrices
        GLCheck(glMatrixMode(GL_PROJECTION));
        GLCheck(glPushMatrix());
        GLCheck(glMatrixMode(GL_MODELVIEW));
        GLCheck(glPushMatrix());

        // Save current render states
        GLCheck(glPushAttrib(GL_ALL_ATTRIB_BITS));

        // Set OpenGL mode flag
        myOpenGLMode = true;
    }
}


////////////////////////////////////////////////////////////
/// End custom OpenGL rendering
////////////////////////////////////////////////////////////
void RenderWindow::EndOpenGL()
{
    if (myOpenGLMode && Window::SetCurrent())
    {
        // Restore render states
        GLCheck(glPopAttrib());

        // Restore the matrices
        GLCheck(glMatrixMode(GL_MODELVIEW));
        GLCheck(glPopMatrix());
        GLCheck(glMatrixMode(GL_PROJECTION));
        GLCheck(glPopMatrix());

        // Clear OpenGL mode flag
        myOpenGLMode = false;
    }
}


////////////////////////////////////////////////////////////
/// Initialize internal window
////////////////////////////////////////////////////////////
void RenderWindow::Initialize()
{
    if (Window::SetCurrent())
    {
        // Set default OpenGL states
        GLCheck(glEnable(GL_ALPHA_TEST));
        GLCheck(glAlphaFunc(GL_GREATER, 0));
        GLCheck(glEnable(GL_BLEND));
        GLCheck(glEnable(GL_TEXTURE_2D));
        GLCheck(glDisable(GL_LIGHTING));
        GLCheck(glDisable(GL_DEPTH_TEST));
        GLCheck(glDisable(GL_CULL_FACE));
        GLCheck(glShadeModel(GL_SMOOTH));
        GLCheck(glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST));
        GLCheck(glClearDepth(1.f));
        GLCheck(glClearStencil(0));

        // Setup the default view
        myDefaultView.Left   = 0;
        myDefaultView.Top    = 0;
        myDefaultView.Width  = static_cast<float>(GetWidth());
        myDefaultView.Height = static_cast<float>(GetHeight());
        myDefaultView.Zoom   = 1.f;
        SetView(NULL);
    }
}

} // namespace sf
