////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Network/SocketUDP.hpp>
#include <SFML/Network/IPAddress.hpp>
#include <SFML/Network/Packet.hpp>
#include <iostream>


namespace sf
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
SocketUDP::SocketUDP() :
mySocket(0),
myPort  (0)
{
    Create();
}


////////////////////////////////////////////////////////////
/// Send an array of bytes
////////////////////////////////////////////////////////////
bool SocketUDP::Send(const char* Data, std::size_t Size, const IPAddress& Address, unsigned short Port)
{
    // First check that socket is valid
    if (mySocket == INVALID_SOCKET)
        return false;

    // Check parameters
    if (Data && Size)
    {
        // Build the target address
        sockaddr_in Target;
        Target.sin_family      = AF_INET;
        Target.sin_port        = htons(Port);
        Target.sin_addr.s_addr = inet_addr(Address.ToString().c_str());
        memset(Target.sin_zero, 0, sizeof(Target.sin_zero));

        // Loop until every byte has been sent
        int Sent = 0;
        int SizeToSend = static_cast<int>(Size);
        for (int Length = 0; Length < SizeToSend; Length += Sent)
        {
            // Send a chunk of data
            Sent = sendto(mySocket, Data + Length, SizeToSend - Length, 0, reinterpret_cast<sockaddr*>(&Target), sizeof(Target));

            // Check errors
            if (Sent <= 0)
                return false;
        }
    }
    else
    {
        // Error...
        std::cerr << "Cannot send data over the network (invalid parameters)" << std::endl;
        return false;
    }

    return true;
}


////////////////////////////////////////////////////////////
/// Receive an array of bytes.
/// This function is blocking, ie. it won't return before some
/// bytes have been received
////////////////////////////////////////////////////////////
bool SocketUDP::Receive(char* Data, std::size_t MaxSize, std::size_t& SizeReceived, IPAddress& Address, unsigned short Port)
{
    // First clear the size received
    SizeReceived = 0;

    // Check that socket is valid
    if (mySocket == INVALID_SOCKET)
        return false;

    // Check parameters
    if (Data && MaxSize)
    {
        // Check if the socket is bound to the specified port
        if (myPort != Port)
        {
            // If the socket was previously bound to another port, we need to recreate it
            if (myPort != 0)
            {
                // Close the socket to unbind it from the port
                Close();

                // Recreate it
                Create();
            }

            // Build an address with the specified port
            sockaddr_in Addr;
            Addr.sin_family      = AF_INET;
            Addr.sin_port        = htons(Port);
            Addr.sin_addr.s_addr = INADDR_ANY;
            memset(Addr.sin_zero, 0, sizeof(Addr.sin_zero));

            // Bind the socket to the port
            if (bind(mySocket, reinterpret_cast<sockaddr*>(&Addr), sizeof(Addr)) == -1)
            {
                std::cerr << "Failed to bind the socket to port " << Port << std::endl;
                myPort = 0;
                return false;
            }

            // Save the new port
            myPort = Port;
        }

        // Data that will be filled with the other computer's address
        sockaddr_in Sender;
        Sender.sin_family      = AF_INET;
        Sender.sin_port        = htons(myPort);
        Sender.sin_addr.s_addr = INADDR_ANY;
        memset(Sender.sin_zero, 0, sizeof(Sender.sin_zero));
        socklen_t SenderSize = sizeof(Sender);

        // Receive a chunk of bytes
        int Received = recvfrom(mySocket, Data, static_cast<int>(MaxSize), 0, reinterpret_cast<sockaddr*>(&Sender), &SenderSize);

        // Check the number of bytes received
        if (Received > 0)
        {
            Address = IPAddress(inet_ntoa(Sender.sin_addr));
            SizeReceived = static_cast<std::size_t>(Received);
        }
        else
        {
            Address = IPAddress();
            return false;
        }
    }
    else
    {
        // Error...
        std::cerr << "Cannot receive data from the network (invalid parameters)" << std::endl;
        return false;
    }

    return true;
}


////////////////////////////////////////////////////////////
/// Send a packet of data
////////////////////////////////////////////////////////////
bool SocketUDP::Send(Packet& PacketToSend, const IPAddress& Address, unsigned short Port)
{
    // Let the packet do custom stuff before sending it
    PacketToSend.OnSend();

    // First send the packet size
    Uint32 PacketSize = htonl(PacketToSend.GetDataSize());
    if (!Send(reinterpret_cast<const char*>(&PacketSize), sizeof(PacketSize), Address, Port))
        return false;

    // Send the packet data
    if (!Send(PacketToSend.GetData(), PacketToSend.GetDataSize(), Address, Port))
        return false;

    return true;
}


////////////////////////////////////////////////////////////
/// Receive a packet.
/// This function is blocking, ie. it won't return before a
/// packet is received
////////////////////////////////////////////////////////////
bool SocketUDP::Receive(Packet& PacketToReceive, IPAddress& Address, unsigned short Port)
{
    // This is not safe at all, as data can be lost, duplicated, or arrive in a different order.
    // So if a packet is split into more than one chunk, nobody knows what could happen...
    // Conclusion : we cannot use packet with UDP, unless we build a more complex protocol on top of it.

    // We start by getting the size of the incoming packet
    std::size_t Received   = 0;
    Uint32      PacketSize = 0;
    Receive(reinterpret_cast<char*>(&PacketSize), sizeof(PacketSize), Received, Address, Port);
    PacketSize = ntohl(PacketSize);

    // Clear the packet
    PacketToReceive.Clear();

    // Use another address instance for receiving the packet data ;
    // chunks of data coming from a different sender will be discarded (and lost...)
    IPAddress Sender;

    // Then loop until we receive all the packet data
    char Buffer[1024];
    while (PacketToReceive.GetDataSize() < PacketSize)
    {
        // Receive a chunk of data
        if (!Receive(Buffer, sizeof(Buffer), Received, Sender, Port))
            return false;

        // Append it into the packet
        if (Sender == Address)
            PacketToReceive.Append(Buffer, Received);
    }

    // Let the packet do custom stuff after data reception
    PacketToReceive.OnReceive();

    return true;
}


////////////////////////////////////////////////////////////
/// Close the socket
////////////////////////////////////////////////////////////
bool SocketUDP::Close()
{
    if (priv::close(mySocket) == -1)
    {
        std::cerr << "Failed to close socket" << std::endl;
        return false;
    }

    return true;
}


////////////////////////////////////////////////////////////
/// Comparison operator ==
////////////////////////////////////////////////////////////
bool SocketUDP::operator ==(const SocketUDP& Other) const
{
    return mySocket == Other.mySocket;
}


////////////////////////////////////////////////////////////
/// Comparison operator !=
////////////////////////////////////////////////////////////
bool SocketUDP::operator !=(const SocketUDP& Other) const
{
    return mySocket != Other.mySocket;
}


////////////////////////////////////////////////////////////
/// Comparison operator <.
/// Provided for compatibility with standard containers, as
/// comparing two sockets doesn't make much sense...
////////////////////////////////////////////////////////////
bool SocketUDP::operator <(const SocketUDP& Other) const
{
    return mySocket < Other.mySocket;
}


////////////////////////////////////////////////////////////
/// Construct the socket from a socket descriptor
/// (for internal use only)
////////////////////////////////////////////////////////////
SocketUDP::SocketUDP(priv::SocketType Descriptor) :
mySocket(Descriptor),
myPort  (0)
{

}


////////////////////////////////////////////////////////////
/// Create the socket
////////////////////////////////////////////////////////////
void SocketUDP::Create()
{
    // Create the socket
    mySocket = socket(PF_INET, SOCK_DGRAM, 0);

    // Clear the last port used
    myPort = 0;

    // To avoid the "Address already in use" error message when trying to bind to the same port
    char Yes = 1;
    if (setsockopt(mySocket, SOL_SOCKET, SO_REUSEADDR, &Yes, sizeof(int)) == -1)
    {
        std::cerr << "Failed to set socket option \"reuse address\" ; "
                  << "binding to a same port may fail if too fast" << std::endl;
    }

    // Enable broadcast by default
    if (setsockopt(mySocket, SOL_SOCKET, SO_BROADCAST, &Yes, sizeof(int)) == -1)
    {
        std::cerr << "Failed to enable broadcast on UDP socket" << std::endl;
    }
}

} // namespace sf
