////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Input.hpp>


namespace sf
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
Input::Input() :
myMouseX(0),
myMouseY(0)
{
    for (int i = 0; i < Key::Count; ++i)
        myKeys[i] = false;

    for (int i = 0; i < 3; ++i)
        myMouseButtons[i] = false;

    myJoystickX[0] = 0;
    myJoystickX[1] = 0;
    myJoystickY[0] = 0;
    myJoystickY[1] = 0;
    myJoystickZ[0] = 0;
    myJoystickZ[1] = 0;

    for (int i = 0; i < 8; ++i)
    {
        myJoystickButtons[0][i] = false;
        myJoystickButtons[1][i] = false;
    }
}


////////////////////////////////////////////////////////////
/// Get the state of a key
////////////////////////////////////////////////////////////
bool Input::IsKeyDown(Key::Code KeyCode) const
{
    return myKeys[KeyCode];
}


////////////////////////////////////////////////////////////
/// Get the state of a mouse button
////////////////////////////////////////////////////////////
bool Input::IsMouseButtonDown(Mouse::Button Button) const
{
    switch (Button)
    {
        case Mouse::Left :   return myMouseButtons[0];
        case Mouse::Right :  return myMouseButtons[1];
        case Mouse::Middle : return myMouseButtons[2];
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Get the state of a joystick button
////////////////////////////////////////////////////////////
bool Input::IsJoystickButtonDown(unsigned int JoyId, unsigned int Button) const
{
    if ((JoyId < 2) && (Button < 8))
        return myJoystickButtons[JoyId][Button];
    else
        return false;
}


////////////////////////////////////////////////////////////
/// Get the mouse left position
////////////////////////////////////////////////////////////
unsigned int Input::GetMouseX() const
{
    return myMouseX;
}


////////////////////////////////////////////////////////////
/// Get the mouse top position
////////////////////////////////////////////////////////////
unsigned int Input::GetMouseY() const
{
    return myMouseY;
}


////////////////////////////////////////////////////////////
/// Get the joystick X position
////////////////////////////////////////////////////////////
int Input::GetJoystickX(unsigned int JoyId) const
{
    return myJoystickX[JoyId];
}


////////////////////////////////////////////////////////////
/// Get the joystick Y position
////////////////////////////////////////////////////////////
int Input::GetJoystickY(unsigned int JoyId) const
{
    return myJoystickY[JoyId];
}


////////////////////////////////////////////////////////////
/// Get the joystick Z position
////////////////////////////////////////////////////////////
int Input::GetJoystickZ(unsigned int JoyId) const
{
    return myJoystickZ[JoyId];
}


////////////////////////////////////////////////////////////
/// /see sfWindowListener::OnEvent
////////////////////////////////////////////////////////////
void Input::OnEvent(const Event& EventReceived)
{
    switch (EventReceived.Type)
    {
        // Key events
        case Event::KeyPressed :  myKeys[EventReceived.Key.Code] = true;  break;
        case Event::KeyReleased : myKeys[EventReceived.Key.Code] = false; break;

        // Mouse pressed event
        case Event::MouseButtonPressed :
            switch (EventReceived.Mouse.Buttons)
            {
                case Mouse::Left :   myMouseButtons[0] = true; break;
                case Mouse::Right :  myMouseButtons[1] = true; break;
                case Mouse::Middle : myMouseButtons[2] = true; break;
            }
            break;

        // Mouse released event
        case Event::MouseButtonReleased :
            switch (EventReceived.Mouse.Buttons)
            {
                case Mouse::Left :   myMouseButtons[0] = false; break;
                case Mouse::Right :  myMouseButtons[1] = false; break;
                case Mouse::Middle : myMouseButtons[2] = false; break;
            }
            break;

        // Mouse move event
        case Event::MouseMove :
            myMouseX = EventReceived.Mouse.X;
            myMouseY = EventReceived.Mouse.Y;
            break;

        // Joystick button events
        case Event::JoystickButtonPressed :  myJoystickButtons[EventReceived.Joystick.JoystickId][EventReceived.Joystick.Button] = true;  break;
        case Event::JoystickButtonReleased : myJoystickButtons[EventReceived.Joystick.JoystickId][EventReceived.Joystick.Button] = false; break;

        // Joystick move event
        case Event::JoystickMove :
            myJoystickX[EventReceived.Joystick.JoystickId] = EventReceived.Joystick.X;
            myJoystickY[EventReceived.Joystick.JoystickId] = EventReceived.Joystick.Y;
            myJoystickZ[EventReceived.Joystick.JoystickId] = EventReceived.Joystick.Z;
            break;

        default :
            break;
    }
}

} // namespace sf
