////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Linux/WindowImplX11.hpp>
#include <SFML/Window/glext/glxext.h>
#include <SFML/Window/glext/glext.h>
#include <X11/keysym.h>
#include <X11/extensions/Xrandr.h>
#include <linux/joystick.h>
#include <fcntl.h>
#include <iostream>
#include <sstream>


namespace sf
{
namespace priv
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
Display*       WindowImplX11::ourDisplay          = NULL;
int            WindowImplX11::ourScreen           = 0;
WindowImplX11* WindowImplX11::ourDummyWindow      = NULL;
WindowImplX11* WindowImplX11::ourFullscreenWindow = NULL;
unsigned int   WindowImplX11::ourWindowsCount     = 0;
unsigned long  WindowImplX11::ourEventMask        = FocusChangeMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                                                    PointerMotionMask | KeyPressMask | KeyReleaseMask | StructureNotifyMask;

////////////////////////////////////////////////////////////
/// Default constructor
/// (creates a dummy window to provide a valid OpenGL context)
////////////////////////////////////////////////////////////
WindowImplX11::WindowImplX11() :
myWindow      (0),
myIsExternal  (false),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Use small dimensions
    myWidth  = 1;
    myHeight = 1;

    // Create the rendering context
    XVisualInfo Visual;
    if (!CreateContext(VideoMode(myWidth, myHeight, 32), false, Visual, 0))
        return;

    // Create a new color map with the chosen visual
    Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);

    // Define the window attributes
    XSetWindowAttributes Attributes;
    Attributes.colormap = ColMap;

    // Create a dummy window (disabled and hidden)
    myWindow = XCreateWindow(ourDisplay,
                             RootWindow(ourDisplay, ourScreen),
                             0, 0,
                             myWidth, myHeight,
                             0,
                             Visual.depth,
                             InputOutput,
                             Visual.visual,
                             CWColormap, &Attributes);

    // Save as the dummy window
    if (!ourDummyWindow)
        ourDummyWindow = this;

    // No joysticks
    myJoysticks[0] = myJoysticks[1] = -1;

    // Set our context as the current OpenGL context for rendering
    SetCurrent();
}


////////////////////////////////////////////////////////////
/// Create the window implementation from an existing control
////////////////////////////////////////////////////////////
WindowImplX11::WindowImplX11(WindowHandle Handle, int AntialiasingLevel) :
myWindow      (0),
myIsExternal  (true),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Save the window handle
    myWindow = Handle;

    if (myWindow)
    {
        // Get the window size
        XWindowAttributes WindowAttributes;
        if (XGetWindowAttributes(ourDisplay, myWindow, &WindowAttributes) == 0)
        {
            std::cerr << "Failed to get the window attributes" << std::endl;
            return;
        }
        myWidth  = WindowAttributes.width;
        myHeight = WindowAttributes.height;

        // Setup the visual infos to match
        XVisualInfo Template;
        Template.depth     = WindowAttributes.depth;
        Template.visualid  = XVisualIDFromVisual(WindowAttributes.visual);
        unsigned long Mask = VisualDepthMask | VisualIDMask;

        // Create the rendering context
        VideoMode Mode = VideoMode::GetDesktopMode();
        Mode.Width  = myWidth;
        Mode.Height = myHeight;
        XVisualInfo Visual;
        if (!CreateContext(Mode, false, Visual, AntialiasingLevel, Template, Mask))
            return;

        // Create a new color map with the chosen visual
        Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);
        XSetWindowColormap(ourDisplay, myWindow, ColMap);

        // Make sure the window is listening to all the requiered events
        XSelectInput(ourDisplay, myWindow, ourEventMask & ~ButtonPressMask);

        // Do some common initializations
        Initialize();
    }
}


////////////////////////////////////////////////////////////
/// Create the window implementation
////////////////////////////////////////////////////////////
WindowImplX11::WindowImplX11(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, int AntialiasingLevel) :
myWindow      (0),
myIsExternal  (false),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Compute position and size
    int Left, Top;
    if (WindowStyle != Fullscreen)
    {
        Left = (DisplayWidth(ourDisplay, ourScreen)  - Mode.Width)  / 2;
        Top  = (DisplayHeight(ourDisplay, ourScreen) - Mode.Height) / 2;
    }
    else
    {
        Left = 0;
        Top  = 0;
    }
    int Width  = myWidth  = Mode.Width;
    int Height = myHeight = Mode.Height;

    // Create the rendering context
    XVisualInfo Visual;
    if (!CreateContext(Mode, WindowStyle == Fullscreen, Visual, AntialiasingLevel))
        return;

    // Create a new color map with the chosen visual
    Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);

    // Define the window attributes
    XSetWindowAttributes Attributes;
    Attributes.event_mask        = ourEventMask;
    Attributes.colormap          = ColMap;
    Attributes.override_redirect = (WindowStyle == Fullscreen);

    // Create the window
    myWindow = XCreateWindow(ourDisplay,
                             RootWindow(ourDisplay, ourScreen),
                             Left, Top,
                             Width, Height,
                             0,
                             Visual.depth,
                             InputOutput,
                             Visual.visual,
                             CWEventMask | CWColormap | CWOverrideRedirect, &Attributes);
    if (!myWindow)
    {
        std::cerr << "Failed to create window" << std::endl;
        return;
    }

    // Set the window's name
    XStoreName(ourDisplay, myWindow, Title.c_str());

    // Set the window's style (tell the windows manager to change our window's decorations and functions according to the requested style)
    Atom WMHintsAtom = XInternAtom(ourDisplay, "_MOTIF_WM_HINTS", false);
    if (WMHintsAtom)
    {
        static const unsigned long MWM_HINTS_FUNCTIONS   = 1 << 0;
        static const unsigned long MWM_HINTS_DECORATIONS = 1 << 1;

        static const unsigned long MWM_DECOR_ALL         = 1 << 0;
        static const unsigned long MWM_DECOR_BORDER      = 1 << 1;
        //static const unsigned long MWM_DECOR_RESIZEH     = 1 << 2;
        static const unsigned long MWM_DECOR_TITLE       = 1 << 3;
        //static const unsigned long MWM_DECOR_MENU        = 1 << 4;
        static const unsigned long MWM_DECOR_MINIMIZE    = 1 << 5;
        //static const unsigned long MWM_DECOR_MAXIMIZE    = 1 << 6;
        
        static const unsigned long MWM_FUNC_ALL          = 1 << 0;
        //static const unsigned long MWM_FUNC_RESIZE       = 1 << 1;
        static const unsigned long MWM_FUNC_MOVE         = 1 << 2;
        static const unsigned long MWM_FUNC_MINIMIZE     = 1 << 3;
        //static const unsigned long MWM_FUNC_MAXIMIZE     = 1 << 4;
        //static const unsigned long MWM_FUNC_CLOSE        = 1 << 5;

        struct WMHints
        {
            unsigned long Flags;
            unsigned long Functions;
            unsigned long Decorations;
            long          InputMode;
            unsigned long State;
        };

        WMHints Hints;
        Hints.Flags = MWM_HINTS_FUNCTIONS | MWM_HINTS_DECORATIONS;

        if (WindowStyle == Resizable)
        {
            Hints.Decorations = MWM_DECOR_ALL;
            Hints.Functions   = MWM_FUNC_ALL;
        }
        else
        {
            Hints.Decorations = MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_MINIMIZE;
            Hints.Functions   = MWM_FUNC_MOVE | MWM_FUNC_MINIMIZE;
        }

        XChangeProperty(ourDisplay, myWindow, WMHintsAtom, WMHintsAtom, 32, PropModeReplace, reinterpret_cast<const unsigned char*>(&Hints), 5);
    }

    // This is a hack to force some windows managers to disable resizing
    if (WindowStyle == Fixed)
    {
        XSizeHints XSizeHints;
        XSizeHints.flags      = PMinSize | PMaxSize;
        XSizeHints.min_width  = XSizeHints.max_width  = Width;
        XSizeHints.min_height = XSizeHints.max_height = Height;
        XSetWMNormalHints(ourDisplay, myWindow, &XSizeHints); 
    }

    // Do some common initializations
    Initialize();

    // In fullscreen mode, we must grab keyboard and mouse inputs
    if (WindowStyle == Fullscreen)
    {
        XGrabPointer(ourDisplay, myWindow, true, 0, GrabModeAsync, GrabModeAsync, myWindow, None, CurrentTime);
        XGrabKeyboard(ourDisplay, myWindow, true, GrabModeAsync, GrabModeAsync, CurrentTime);
    }
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
WindowImplX11::~WindowImplX11()
{
    // Cleanup graphical resources
    CleanUp();

    // Destroy the window
    if (myWindow && !myIsExternal)
        XDestroyWindow(ourDisplay, myWindow);

    // Close the display
    CloseDisplay();
}


////////////////////////////////////////////////////////////
/// /see WindowImpl::Display
////////////////////////////////////////////////////////////
void WindowImplX11::Display()
{
    if (myWindow && myGLContext)
        glXSwapBuffers(ourDisplay, myWindow);
}


////////////////////////////////////////////////////////////
/// /see WindowImpl::ProcessEvents
////////////////////////////////////////////////////////////
void WindowImplX11::ProcessEvents()
{
    // Get the number of events in the queue
    int NbEvents = XPending(ourDisplay);

    // Process the events
    XEvent Event;
    for (int i = 0; i < NbEvents; ++i)
    {
        XNextEvent(ourDisplay, &Event);
        if (Event.xany.window == myWindow)
            ProcessEvent(Event);
        else
            XPutBackEvent(ourDisplay, &Event);
    }

    // Read the joysticks state
    if (myJoysticks[0] > 0) ReadJoyState(0);
    if (myJoysticks[1] > 0) ReadJoyState(1);
}


////////////////////////////////////////////////////////////
/// /see WindowImpl::MakeCurrent
////////////////////////////////////////////////////////////
void WindowImplX11::MakeCurrent() const
{
    if (myWindow && myGLContext)
        glXMakeCurrent(ourDisplay, myWindow, myGLContext);
}


////////////////////////////////////////////////////////////
/// /see WindowImpl::UseVerticalSync
////////////////////////////////////////////////////////////
void WindowImplX11::UseVerticalSync(bool Enabled)
{
    const GLubyte* ProcAddress = reinterpret_cast<const GLubyte*>("glXSwapIntervalSGI");
    PFNGLXSWAPINTERVALSGIPROC glXSwapIntervalSGI = reinterpret_cast<PFNGLXSWAPINTERVALSGIPROC>(glXGetProcAddress(ProcAddress));
    if (glXSwapIntervalSGI)
        glXSwapIntervalSGI(Enabled ? 1 : 0);
}


////////////////////////////////////////////////////////////
/// /see WindowImpl::ShowMouseCursor
////////////////////////////////////////////////////////////
void WindowImplX11::ShowMouseCursor(bool Show)
{
    XDefineCursor(ourDisplay, myWindow, Show ? None : myHiddenCursor);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::SetPosition
////////////////////////////////////////////////////////////
void WindowImplX11::SetPosition(int Left, int Top)
{
    XMoveWindow(ourDisplay, myWindow, Left, Top);
}


////////////////////////////////////////////////////////////
/// Create the OpenGL rendering context
////////////////////////////////////////////////////////////
bool WindowImplX11::CreateContext(VideoMode Mode, bool Fullscreen, XVisualInfo& ChosenVisual, int AntialiasingLevel, XVisualInfo Template, unsigned long Mask)
{
    // Fullscreen mode
    if (Fullscreen)
    {
        // Check if the XRandR extension is present
        int Version;
        if (XQueryExtension(ourDisplay, "RANDR", &Version, &Version, &Version))
        {
            // Get the current configuration
            XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
            if (Config)
            {
                // Get the current rotation
                Rotation CurrentRotation;
                myOldVideoMode = XRRConfigCurrentConfiguration(Config, &CurrentRotation);

                // Get the available screen sizes
                int NbSizes;
                XRRScreenSize* Sizes = XRRConfigSizes(Config, &NbSizes);
                if (Sizes && (NbSizes > 0))
                {
                    // Search a matching size
                    for (int i = 0; i < NbSizes; ++i)
                    {
                        if ((Sizes[i].width == static_cast<int>(Mode.Width)) && (Sizes[i].height == static_cast<int>(Mode.Height)))
                        {
                            // Switch to fullscreen mode
                            XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), i, CurrentRotation, CurrentTime);

                            // Set "this" as the current fullscreen window
                            ourFullscreenWindow = this;
                            break;
                        }
                    }
                }

                // Free the configuration instance
                XRRFreeScreenConfigInfo(Config);
            }
            else
            {
                // Failed to get the screen configuration
                std::cerr << "Failed to get the current screen configuration for fullscreen mode, switching to windiw mode" << std::endl;
            }
        }
        else
        {
            // XRandr extension is not supported : we cannot use fullscreen mode
            std::cerr << "Fullscreen is not supported, switching to window mode" << std::endl;
        }
    }

    // Get all the visuals matching the template
    Template.screen = ourScreen;
    int NbVisuals = 0;
    XVisualInfo* Visuals = XGetVisualInfo(ourDisplay, Mask | VisualScreenMask, &Template, &NbVisuals);
    if (!Visuals || (NbVisuals == 0))
    {
        if (Visuals)
            XFree(Visuals);
        std::cerr << "There is no valid visual for the selected screen" << std::endl;
        return false;
    }

    // A helper structure for storing visual attributes
    struct VisualAttributes
    {
        int RGBA, DoubleBuffer, Red, Green, Blue, Alpha, Depth, Stencil, MultiSampling, Samples;
    };

    // Find the best visual
    bool Found = false;
    XVisualInfo* Visual = &Visuals[0];
    while (!Found)
    {
        VisualAttributes BestAttribs = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
        int MultiSampling = (AntialiasingLevel > 0 ? 1 : 0);

        for (int i = 0; i < NbVisuals; ++i)
        {
            // Get the current visual attributes
            VisualAttributes Attribs = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
            glXGetConfig(ourDisplay, &Visuals[i], GLX_RGBA,               &Attribs.RGBA);
            glXGetConfig(ourDisplay, &Visuals[i], GLX_DOUBLEBUFFER,       &Attribs.DoubleBuffer); 
            glXGetConfig(ourDisplay, &Visuals[i], GLX_RED_SIZE,           &Attribs.Red);
            glXGetConfig(ourDisplay, &Visuals[i], GLX_GREEN_SIZE,         &Attribs.Green); 
            glXGetConfig(ourDisplay, &Visuals[i], GLX_BLUE_SIZE,          &Attribs.Blue); 
            glXGetConfig(ourDisplay, &Visuals[i], GLX_ALPHA_SIZE,         &Attribs.Alpha); 
            glXGetConfig(ourDisplay, &Visuals[i], GLX_DEPTH_SIZE,         &Attribs.Depth);        
            glXGetConfig(ourDisplay, &Visuals[i], GLX_STENCIL_SIZE,       &Attribs.Stencil);
            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLE_BUFFERS_ARB, &Attribs.MultiSampling);        
            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLES_ARB,        &Attribs.Samples);
    
            // Compare with the current best
            if (((AntialiasingLevel == 0) || (Attribs.MultiSampling == MultiSampling) && (Attribs.Samples == AntialiasingLevel)) &&
               ((BestAttribs.RGBA    < Attribs.RGBA)    || (BestAttribs.DoubleBuffer < Attribs.DoubleBuffer) ||
                (BestAttribs.Red     < Attribs.Red)     || (BestAttribs.Green        < Attribs.Green)        ||
                (BestAttribs.Blue    < Attribs.Blue)    || (BestAttribs.Alpha        < Attribs.Alpha)        ||
                (BestAttribs.Stencil < Attribs.Stencil) || (BestAttribs.Depth        < Attribs.Depth)))
            {
                BestAttribs = Attribs;
                Visual      = &Visuals[i];
                Found       = true;
            }
        }

        // If no visual has been found, try a lower level of antialiasing
        if (!Found)
        {
            if (AntialiasingLevel > 2)
            {
                std::cerr << "Failed to find a pixel format supporting "
                          << AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;
                AntialiasingLevel = 2;
            }
            else if (AntialiasingLevel > 0)
            {
                std::cerr << "Failed to find a pixel format supporting antialiasing ; antialiasing will be disabled" << std::endl;
                AntialiasingLevel = 0;
            }
            else
            {
                std::cerr << "Failed to find a suitable pixel format for the window -- cannot create OpenGL context" << std::endl;
                return false;
            }
        }
    }

    // Create the OpenGL context
    myGLContext = glXCreateContext(ourDisplay, Visual, glXGetCurrentContext(), true);
    if (myGLContext == NULL)
    {
        std::cerr << "Failed to create an OpenGL context for this window" << std::endl;
        return false;
    }

    // Extract the depth and stencil bits from the chosen format
    int Depth, Stencil;
    glXGetConfig(ourDisplay, Visual, GLX_DEPTH_SIZE,   &Depth);
    glXGetConfig(ourDisplay, Visual, GLX_STENCIL_SIZE, &Stencil);
    myDepthBits   = static_cast<unsigned int>(Depth);
    myStencilBits = static_cast<unsigned int>(Stencil);

    // Assign the chosen visual, and free the temporary visuals array
    ChosenVisual = *Visual;
    XFree(Visuals);
    
    // Enable multisampling if needed
    if (AntialiasingLevel > 0)
        glEnable(GL_MULTISAMPLE_ARB);

    return true;
}


////////////////////////////////////////////////////////////
/// Do some common initializations after the window has been created
////////////////////////////////////////////////////////////
void WindowImplX11::Initialize()
{
    // Get the atom defining the close event
    myAtomClose = XInternAtom(ourDisplay, "WM_DELETE_WINDOW", false);
    XSetWMProtocols(ourDisplay, myWindow, &myAtomClose, 1);

    // Show the window
    XMapWindow(ourDisplay, myWindow);
    XFlush(ourDisplay);

    // Setup the joysticks
    SetupJoystick(0);
    SetupJoystick(1);

    // Create the hiden cursor
    CreateHiddenCursor();

    // Set our context as the current OpenGL context for rendering
    SetCurrent();

    // Flush the commands queue
    XFlush(ourDisplay);
}


////////////////////////////////////////////////////////////
/// Setup a joystick
////////////////////////////////////////////////////////////
void WindowImplX11::SetupJoystick(unsigned int Id)
{
    // Open the joystick handle
    std::ostringstream oss;
    oss << "/dev/input/js" << Id;
    myJoysticks[Id] = open(oss.str().c_str(), O_RDONLY);
    if (myJoysticks[Id] > 0)
    {
        // Use non-blocking mode
        fcntl(myJoysticks[Id], F_SETFL, O_NONBLOCK);
    }
}


////////////////////////////////////////////////////////////
/// Read the state of a given joystick, and triggers the
/// appropriate events
////////////////////////////////////////////////////////////
void WindowImplX11::ReadJoyState(unsigned int Id)
{
    js_event JoyState;
    while (read(myJoysticks[Id], &JoyState, sizeof(JoyState)) > 0)
    {
        switch (JoyState.type & ~JS_EVENT_INIT)
        {
            // An axis has been moved
            case JS_EVENT_AXIS :
            {
                Event Evt;
                Evt.Type                = Event::JoystickMove;
                Evt.Joystick.JoystickId = Id;
                if      (JoyState.number == 0) Evt.Joystick.X = JoyState.value * 100 / 32767;
                else if (JoyState.number == 1) Evt.Joystick.Y = JoyState.value * 100 / 32767;
                else if (JoyState.number == 2) Evt.Joystick.Z = JoyState.value * 100 / 32767;

                SendEvent(Evt);
                break;
            }

            // A button has been pressed
            case JS_EVENT_BUTTON :
            {
                Event Evt;
                Evt.Type                = JoyState.value ? Event::JoystickButtonPressed : Event::JoystickButtonReleased;
                Evt.Joystick.JoystickId = Id;
                Evt.Joystick.Button     = JoyState.number;

                SendEvent(Evt);
                break;
            }
        }
    }
}


////////////////////////////////////////////////////////////
/// Create a transparent mouse cursor
////////////////////////////////////////////////////////////
void WindowImplX11::CreateHiddenCursor()
{
    // Create the cursor's pixmap (1x1 pixels)
	Pixmap CursorPixmap = XCreatePixmap(ourDisplay, myWindow, 1, 1, 1);
    GC GraphicsContext = XCreateGC(ourDisplay, CursorPixmap, 0, NULL);
    XDrawPoint(ourDisplay, CursorPixmap, GraphicsContext, 0, 0);
    XFreeGC(ourDisplay, GraphicsContext);

    // Create the cursor, using the pixmap as both the shape and the mask of the cursor
	XColor Color;
	Color.flags = DoRed | DoGreen | DoBlue;
	Color.red = Color.blue = Color.green = 0;
    myHiddenCursor = XCreatePixmapCursor(ourDisplay, CursorPixmap, CursorPixmap, &Color, &Color, 0, 0);

    // We don't need the pixmap any longer, free it
	XFreePixmap(ourDisplay, CursorPixmap);
}


////////////////////////////////////////////////////////////
/// Cleanup graphical resources attached to the window
////////////////////////////////////////////////////////////
void WindowImplX11::CleanUp()
{
    // Restore previous video mode (in case we were running in fullscreen)
    if (ourFullscreenWindow == this)
    {
        // Get current screen info
        XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
        if (Config) 
        {
            // Get the current rotation
            Rotation CurrentRotation;
            XRRConfigCurrentConfiguration(Config, &CurrentRotation);

            // Reset the video mode
            XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), myOldVideoMode, CurrentRotation, CurrentTime);

            // Free the configuration instance
            XRRFreeScreenConfigInfo(Config);
        } 

        // Reset the fullscreen window
        ourFullscreenWindow = NULL;
    }

    // Unhide the mouse cursor (in case it was hidden)
    ShowMouseCursor(true);

    // Close the joysticks handles
    if (myJoysticks[0] > 0) close(myJoysticks[0]);
    if (myJoysticks[1] > 0) close(myJoysticks[1]);

    // Destroy the OpenGL context
    if (myGLContext)
    {
        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
        if (this != ourDummyWindow)
        {
            if (ourDummyWindow)
                ourDummyWindow->SetCurrent();
        }
        else
        {
            ourDummyWindow = NULL;
            glXMakeCurrent(ourDisplay, 0, NULL);
        }

        // Now that our context is no more the current one, we can destroy it
        glXDestroyContext(ourDisplay, myGLContext);
        myGLContext = NULL;
    }
}


////////////////////////////////////////////////////////////
/// Process an incoming event from the window
////////////////////////////////////////////////////////////
void WindowImplX11::ProcessEvent(XEvent WinEvent)
{
    switch (WinEvent.type)
    {
        // Destroy event
        case DestroyNotify :
        {
            // The window is about to be destroyed : we must cleanup resources
            CleanUp();
            break;
        }

        // Gain focus event
        case FocusIn :
        {
            Event Evt;
            Evt.Type = Event::GainedFocus;
            SendEvent(Evt);
            break;
        }

        // Lost focus event
        case FocusOut :
        {
            Event Evt;
            Evt.Type = Event::LostFocus;
            SendEvent(Evt);
            break;
        }

        // Resize event
        case ConfigureNotify :
        {
            if ((WinEvent.xconfigure.width != static_cast<int>(myWidth)) || (WinEvent.xconfigure.height != static_cast<int>(myHeight)))
            {
                Event Evt;
                Evt.Type        = Event::Resize;
                Evt.Size.Width  = WinEvent.xconfigure.width;
                Evt.Size.Height = WinEvent.xconfigure.height;
                SendEvent(Evt);
            }
            break;
        }

        // Close event
        case ClientMessage :
        {
            if ((WinEvent.xclient.format == 32) && (WinEvent.xclient.data.l[0]) == static_cast<long>(myAtomClose))  
            {
                Event Evt;
                Evt.Type = Event::Close;
                SendEvent(Evt);
                CleanUp();
            }
            break;
        }

        // Key down event
        case KeyPress :
        {
            // Get the keysym of the key that has been pressed
            char Buffer[32];
            KeySym Sym;
            XLookupString(&WinEvent.xkey, Buffer, 32, &Sym, NULL);

            // Fill the event parameters
            Event Evt;
            Evt.Type        = Event::KeyPressed;
            Evt.Key.Code    = KeysymToSF(Sym);
            Evt.Key.Alt     = WinEvent.xkey.state & Mod1Mask;
            Evt.Key.Control = WinEvent.xkey.state & ControlMask;
            Evt.Key.Shift   = WinEvent.xkey.state & ShiftMask;

            // TODO : Generate a TextEntered event

            SendEvent(Evt);
            break;
        }

        // Key up event
        case KeyRelease :
        {
            // Get the keysym of the key that has been pressed
            char Buffer[32];
            KeySym Sym;
            XLookupString(&WinEvent.xkey, Buffer, 32, &Sym, NULL);

            // Fill the event parameters
            Event Evt;
            Evt.Type        = Event::KeyReleased;
            Evt.Key.Code    = KeysymToSF(Sym);
            Evt.Key.Alt     = WinEvent.xkey.state & Mod1Mask;
            Evt.Key.Control = WinEvent.xkey.state & ControlMask;
            Evt.Key.Shift   = WinEvent.xkey.state & ShiftMask;

            // TODO : Generate a TextEntered event

            SendEvent(Evt);
            break;
        }

        // Mouse button pressed
        case ButtonPress :
        {
            Event Evt;
            Evt.Type    = Event::MouseButtonPressed;
            Evt.Mouse.X = WinEvent.xbutton.x;
            Evt.Mouse.Y = WinEvent.xbutton.y;
            switch (WinEvent.xbutton.button)
            {
                case Button1 : Evt.Mouse.Buttons = Mouse::Left;   break;
                case Button2 : Evt.Mouse.Buttons = Mouse::Middle; break;
                case Button3 : Evt.Mouse.Buttons = Mouse::Right;  break;
            }
            SendEvent(Evt);
            break;
        }

        // Mouse button released
        case ButtonRelease :
        {
            if ((WinEvent.xbutton.button == Button1) || (WinEvent.xbutton.button == Button2) || (WinEvent.xbutton.button == Button3))
            {
                Event Evt;
                Evt.Type    = Event::MouseButtonReleased;
                Evt.Mouse.X = WinEvent.xbutton.x;
                Evt.Mouse.Y = WinEvent.xbutton.y;
                switch (WinEvent.xbutton.button)
                {
                    case Button1 : Evt.Mouse.Buttons = Mouse::Left;   break;
                    case Button2 : Evt.Mouse.Buttons = Mouse::Middle; break;
                    case Button3 : Evt.Mouse.Buttons = Mouse::Right;  break;
                }
                SendEvent(Evt);
            }
            else
            {
                Event Evt;
                Evt.Type             = Event::MouseWheelMoved;
                Evt.MouseWheel.Delta = WinEvent.xbutton.button == Button4 ? 1 : -1;
                SendEvent(Evt);
            }
            break;
        }

        // Mouse moved
        case MotionNotify :
        {
            Event Evt;
            Evt.Type    = Event::MouseMove;
            Evt.Mouse.X = WinEvent.xmotion.x;
            Evt.Mouse.Y = WinEvent.xmotion.y;
            Evt.Mouse.Buttons = 0;
            if (WinEvent.xmotion.state & Button1Mask) Evt.Mouse.Buttons |= Mouse::Left;
            if (WinEvent.xmotion.state & Button2Mask) Evt.Mouse.Buttons |= Mouse::Middle;
            if (WinEvent.xmotion.state & Button3Mask) Evt.Mouse.Buttons |= Mouse::Right;
            SendEvent(Evt);
            break;
        }
    }
}


////////////////////////////////////////////////////////////
/// Convert a X11 keysym to SFML key code
////////////////////////////////////////////////////////////
Key::Code WindowImplX11::KeysymToSF(KeySym Sym)
{
    // First convert to uppercase (to avoid dealing with two different keysyms for the same key)
    KeySym Lower, Key;
    XConvertCase(Sym, &Lower, &Key);

    switch (Key)
    {
        case XK_Escape :      return Key::Escape;
        case XK_space :       return Key::Space;
        case XK_Return :      return Key::Return;
        case XK_BackSpace :   return Key::Back;
        case XK_Tab :         return Key::Tab;
        case XK_Prior :       return Key::PageUp;
        case XK_Next :        return Key::PageDown;
        case XK_End :         return Key::End;
        case XK_Home :        return Key::Home;
        case XK_Insert :      return Key::Insert;
        case XK_Delete :      return Key::Delete;
        case XK_KP_Add :      return Key::Add;
        case XK_KP_Subtract : return Key::Subtract;
        case XK_KP_Multiply : return Key::Multiply;
        case XK_KP_Divide :   return Key::Divide;
        case XK_Pause :       return Key::Pause;
        case XK_F1 :          return Key::F1;
        case XK_F2 :          return Key::F2;
        case XK_F3 :          return Key::F3;
        case XK_F4 :          return Key::F4;
        case XK_F5 :          return Key::F5;
        case XK_F6 :          return Key::F6;
        case XK_F7 :          return Key::F7;
        case XK_F8 :          return Key::F8;
        case XK_F9 :          return Key::F9;
        case XK_F10 :         return Key::F10;
        case XK_F11 :         return Key::F11;
        case XK_F12 :         return Key::F12;
        case XK_F13 :         return Key::F13;
        case XK_F14 :         return Key::F14;
        case XK_F15 :         return Key::F15;
        case XK_Left :        return Key::Left;
        case XK_Right :       return Key::Right;
        case XK_Up :          return Key::Up;
        case XK_Down :        return Key::Down;
        case XK_KP_0 :        return Key::Numpad0;
        case XK_KP_1 :        return Key::Numpad1;
        case XK_KP_2 :        return Key::Numpad2;
        case XK_KP_3 :        return Key::Numpad3;
        case XK_KP_4 :        return Key::Numpad4;
        case XK_KP_5 :        return Key::Numpad5;
        case XK_KP_6 :        return Key::Numpad6;
        case XK_KP_7 :        return Key::Numpad7;
        case XK_KP_8 :        return Key::Numpad8;
        case XK_KP_9 :        return Key::Numpad9;
        case XK_A :           return Key::A;
        case XK_Z :           return Key::Z;
        case XK_E :           return Key::E;
        case XK_R :           return Key::R;
        case XK_T :           return Key::T;
        case XK_Y :           return Key::Y;
        case XK_U :           return Key::U;
        case XK_I :           return Key::I;
        case XK_O :           return Key::O;
        case XK_P :           return Key::P;
        case XK_Q :           return Key::Q;
        case XK_S :           return Key::S;
        case XK_D :           return Key::D;
        case XK_F :           return Key::F;
        case XK_G :           return Key::G;
        case XK_H :           return Key::H;
        case XK_J :           return Key::J;
        case XK_K :           return Key::K;
        case XK_L :           return Key::L;
        case XK_M :           return Key::M;
        case XK_W :           return Key::W;
        case XK_X :           return Key::X;
        case XK_C :           return Key::C;
        case XK_V :           return Key::V;
        case XK_B :           return Key::B;
        case XK_N :           return Key::N;
        case XK_0 :           return Key::Num0;
        case XK_1 :           return Key::Num1;
        case XK_2 :           return Key::Num2;
        case XK_3 :           return Key::Num3;
        case XK_4 :           return Key::Num4;
        case XK_5 :           return Key::Num5;
        case XK_6 :           return Key::Num6;
        case XK_7 :           return Key::Num7;
        case XK_8 :           return Key::Num8;
        case XK_9 :           return Key::Num9;
    }

    return Key::Code(0);
}


////////////////////////////////////////////////////////////
/// Open the display (if not already done)
////////////////////////////////////////////////////////////
bool WindowImplX11::OpenDisplay(bool AddWindow)
{
    // If no display has been opened yet, open it
    if (ourDisplay == NULL)
    {
        ourDisplay = XOpenDisplay(NULL);
        if (ourDisplay)
            ourScreen = DefaultScreen(ourDisplay);
        else
            std::cerr << "Failed to open a connection with the X server" << std::endl;
    }

    // Increase the number of windows
    if (AddWindow)
        ourWindowsCount++;

    return ourDisplay != NULL;
}


////////////////////////////////////////////////////////////
/// Close the display
////////////////////////////////////////////////////////////
void WindowImplX11::CloseDisplay()
{
    // Decrease the number of windows
    ourWindowsCount--;

    // If all windows have been destroyed, then we can close the display
    if (ourWindowsCount == 0)
    {
        XCloseDisplay(ourDisplay);
        ourDisplay = NULL;
    }
}

} // namespace priv

} // namespace sf
