////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#define _WIN32_WINNT 0x500
#include <SFML/Window/Win32/WindowImplWin32.hpp>
#include <GL/gl.h>
#include <SFML/Window/glext/wglext.h>
#include <SFML/Window/glext/glext.h>
#include <mmsystem.h>
#include <iostream>


namespace sf
{
namespace priv
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
unsigned int     WindowImplWin32::ourWindowCount      = 0;
const wchar_t*   WindowImplWin32::ourClassName        = L"SFML_Window";
WindowImplWin32* WindowImplWin32::ourDummyWindow      = NULL;
WindowImplWin32* WindowImplWin32::ourFullscreenWindow = NULL;


////////////////////////////////////////////////////////////
/// Default constructor
/// (creates a dummy window to provide a valid OpenGL context)
////////////////////////////////////////////////////////////
WindowImplWin32::WindowImplWin32() :
myHandle  (NULL),
myCallback(0),
myCursor  (NULL)
{
    // Register the window class at first call
    if (ourWindowCount == 0)
        RegisterWindowClass();

    // Use small dimensions
    myWidth  = 1;
    myHeight = 1;

    // Create a dummy window (disabled and hidden)
    myHandle = CreateWindow(ourClassName, TEXT(""), WS_POPUP | WS_DISABLED, 0, 0, myWidth, myHeight, NULL, NULL, GetModuleHandle(NULL), NULL);
    ShowWindow(myHandle, SW_HIDE);

    // Create the rendering context
    if (myHandle)
        CreateContext(VideoMode(myWidth, myHeight, 32), false, 0);

    // Save as the dummy window
    if (!ourDummyWindow)
        ourDummyWindow = this;
}


////////////////////////////////////////////////////////////
/// Create the window implementation from an existing control
////////////////////////////////////////////////////////////
WindowImplWin32::WindowImplWin32(WindowHandle Handle, int AntialiasingLevel) :
myCallback(0),
myCursor  (NULL)
{
    // Save window handle
    myHandle = static_cast<HWND>(Handle);

    if (myHandle)
    {
        // Get window client size
        RECT Rect;
        GetClientRect(myHandle, &Rect);
        myWidth  = Rect.right - Rect.left;
        myHeight = Rect.bottom - Rect.top;

        // Create the rendering context
        VideoMode Mode = VideoMode::GetDesktopMode();
        Mode.Width  = myWidth;
        Mode.Height = myHeight;
        CreateContext(Mode, false, AntialiasingLevel);

        // We change the event procedure of the control (it is important to save the old one)
        SetWindowLongPtr(myHandle, GWLP_USERDATA, reinterpret_cast<long>(this));
        myCallback = SetWindowLongPtr(myHandle, GWLP_WNDPROC, reinterpret_cast<long>(&WindowImplWin32::GlobalOnEvent));

        // Setup joysticks
        SetupJoysticks();
    }
}


////////////////////////////////////////////////////////////
/// Create the window implementation
////////////////////////////////////////////////////////////
WindowImplWin32::WindowImplWin32(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, int AntialiasingLevel) :
myHandle  (NULL),
myCallback(0),
myCursor  (NULL)
{
    // Register the window class at first call
    if (ourWindowCount == 0)
        RegisterWindowClass();

    // Compute position and size
    int Left   = (GetDeviceCaps(GetDC(NULL), HORZRES) - Mode.Width)  / 2;
    int Top    = (GetDeviceCaps(GetDC(NULL), VERTRES) - Mode.Height) / 2;
    int Width  = myWidth  = Mode.Width;
    int Height = myHeight = Mode.Height;

    // Choose the window style according to the Style parameter
    DWORD Win32Style = WS_VISIBLE;
    Win32Style |= (WindowStyle == Resizable ? WS_OVERLAPPEDWINDOW : WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX);

    // In windowed mode, adjust width and height so that window will have the requested client area
    if (WindowStyle != Fullscreen)
    {
        RECT Rect = {0, 0, Width, Height};
        AdjustWindowRect(&Rect, Win32Style, false);
        Width  = Rect.right - Rect.left;
        Height = Rect.bottom - Rect.top;
    }

    // Create the window
    wchar_t WTitle[256];
    int NbChars = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, Title.c_str(), static_cast<int>(Title.size()), WTitle, sizeof(WTitle) / sizeof(*WTitle));
    WTitle[NbChars] = L'\0';
    myHandle = CreateWindow(ourClassName, WTitle, Win32Style, Left, Top, Width, Height, NULL, NULL, GetModuleHandle(NULL), this);

    // Create the rendering context
    if (myHandle)
        CreateContext(Mode, WindowStyle == Fullscreen, AntialiasingLevel);

    // Increment window count
    ourWindowCount++;

    // Setup joysticks
    SetupJoysticks();
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
WindowImplWin32::~WindowImplWin32()
{
    if (!myCallback)
    {
        // Destroy the window
        if (myHandle)
            DestroyWindow(myHandle);

        // Decrement the window count
        ourWindowCount--;

        // Unregister window class if we were the last window
        if (ourWindowCount == 0)
            UnregisterClass(ourClassName, GetModuleHandle(NULL));
    }
    else
    {
        // The window is external : remove the hook on its message callback
        SetWindowLongPtr(myHandle, GWLP_WNDPROC, myCallback);
    }
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::Display
////////////////////////////////////////////////////////////
void WindowImplWin32::Display()
{
    if (myDeviceContext && myGLContext)
        SwapBuffers(myDeviceContext);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ProcessEvents
////////////////////////////////////////////////////////////
void WindowImplWin32::ProcessEvents()
{
    // We update the window only if we own it
    if (!myCallback)
    {
        MSG Message;
        while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&Message);
            DispatchMessage(&Message);
        }
    }
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::MakeCurrent
////////////////////////////////////////////////////////////
void WindowImplWin32::MakeCurrent() const
{
    if (myDeviceContext && myGLContext)
        wglMakeCurrent(myDeviceContext, myGLContext);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::UseVerticalSync
////////////////////////////////////////////////////////////
void WindowImplWin32::UseVerticalSync(bool Enabled)
{
    PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT = reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(wglGetProcAddress("wglSwapIntervalEXT"));
    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(Enabled ? 1 : 0);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ShowMouseCursor
////////////////////////////////////////////////////////////
void WindowImplWin32::ShowMouseCursor(bool Show)
{
    if (Show)
        myCursor = LoadCursor(NULL, IDC_ARROW);
    else
        myCursor = NULL;

    SetCursor(myCursor);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::SetPosition
////////////////////////////////////////////////////////////
void WindowImplWin32::SetPosition(int Left, int Top)
{
    SetWindowPos(myHandle, NULL, Left, Top, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}


////////////////////////////////////////////////////////////
/// Register the window class
////////////////////////////////////////////////////////////
void WindowImplWin32::RegisterWindowClass()
{
    WNDCLASSW WindowClass;
    WindowClass.style         = 0;
    WindowClass.lpfnWndProc   = &WindowImplWin32::GlobalOnEvent;
    WindowClass.cbClsExtra    = 0;
    WindowClass.cbWndExtra    = 0;
    WindowClass.hInstance     = GetModuleHandle(NULL);
    WindowClass.hIcon         = NULL;
    WindowClass.hCursor       = 0;
    WindowClass.hbrBackground = 0;
    WindowClass.lpszMenuName  = NULL;
    WindowClass.lpszClassName = ourClassName;

    RegisterClass(&WindowClass);
}


////////////////////////////////////////////////////////////
/// Create the OpenGL rendering context
////////////////////////////////////////////////////////////
void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int AntialiasingLevel)
{
    // Fullscreen mode
    if (Fullscreen)
    {
        DEVMODE DevMode;
        DevMode.dmSize       = sizeof(DEVMODE);
        DevMode.dmPelsWidth  = Mode.Width;
        DevMode.dmPelsHeight = Mode.Height;
        DevMode.dmBitsPerPel = Mode.BitsPerPixel;
        DevMode.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

        // Apply fullscreen mode
        if (ChangeDisplaySettings(&DevMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
        {
            std::cerr << "Failed to change display mode for fullscreen -- cannot create OpenGL context" << std::endl;
            return;
        }

        // Change window style (no border, no titlebar, ...)
        SetWindowLong(myHandle, GWL_STYLE,   WS_POPUP);
        SetWindowLong(myHandle, GWL_EXSTYLE, WS_EX_APPWINDOW);

        // And resize it so that it fits the entire screen
        SetWindowPos(myHandle, HWND_TOP, 0, 0, Mode.Width, Mode.Height, SWP_FRAMECHANGED);
        ShowWindow(myHandle, SW_SHOW);

        // Set "this" as the current fullscreen window
        ourFullscreenWindow = this;
    }

    // SetPixelFormat can fail (really ?) if window style doesn't contain these flags
    long Style = GetWindowLong(myHandle, GWL_STYLE);
    SetWindowLong(myHandle, GWL_STYLE, Style | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);

    // Get the device context attached to the window
    myDeviceContext = GetDC(myHandle);
    if (myDeviceContext == NULL)
    {
        std::cerr << "Failed to get device context of window -- cannot create OpenGL context" << std::endl;
        return;
    }

    // Let's find a suitable pixel format -- first try with antialiasing
    int PixelFormat = 0;
    if (AntialiasingLevel > 0)
    {
        // Get the wglChoosePixelFormatARB function (it is an extension)
        PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress("wglChoosePixelFormatARB"));

        // Define the basic attributes we want for our window
        int IntAttributes[] =
        {
		    WGL_SAMPLES_ARB,        AntialiasingLevel,
		    WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,
            WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
		    WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
		    WGL_ACCELERATION_ARB,   WGL_FULL_ACCELERATION_ARB,
		    WGL_COLOR_BITS_ARB,     Mode.BitsPerPixel,
		    WGL_DOUBLE_BUFFER_ARB,  GL_TRUE,
		    0,                      0
        };

        // Let's check how many formats are supporting our requirements
        int   Formats[64];
	    UINT  NbFormats;
	    float FloatAttributes[] = {0, 0};
	    bool  IsValid = wglChoosePixelFormatARB(myDeviceContext, IntAttributes, FloatAttributes, sizeof(Formats) / sizeof(*Formats), Formats, &NbFormats) != 0;
        if (!IsValid || (NbFormats == 0))
        {
            if (AntialiasingLevel > 2)
            {
                // No format matching our needs : reduce the multisampling level
                std::cerr << "Failed to find a pixel format supporting "
                          << AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;

                IntAttributes[1] = 2;
	            IsValid = wglChoosePixelFormatARB(myDeviceContext, IntAttributes, FloatAttributes, sizeof(Formats) / sizeof(*Formats), Formats, &NbFormats) != 0;
            }

            if (!IsValid || (NbFormats == 0))
            {
                // Cannot find any pixel format supporting multisampling ; disabling antialiasing
                std::cerr << "Failed to find a pixel format supporting antialiasing ; antialiasing will be disabled" << std::endl;
                AntialiasingLevel = 0;
            }
        }

        // Get the best format among the returned ones
        if (IsValid && (NbFormats > 0))
        {
            PixelFormat = Formats[0];
            PIXELFORMATDESCRIPTOR BestAttribs;
            BestAttribs.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
            BestAttribs.nVersion = 1;
            DescribePixelFormat(myDeviceContext, PixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &BestAttribs);
            for (UINT i = 1; i < NbFormats; ++i)
            {
                // Get the current format's attributes
                PIXELFORMATDESCRIPTOR Attribs;
                Attribs.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
                Attribs.nVersion = 1;
                DescribePixelFormat(myDeviceContext, Formats[i], sizeof(PIXELFORMATDESCRIPTOR), &Attribs);

                // Compare them with the current best format
                if ((BestAttribs.cStencilBits < Attribs.cStencilBits) || (BestAttribs.cDepthBits < Attribs.cDepthBits))
                {
                    PixelFormat = Formats[i];
                    BestAttribs = Attribs;
                }
            }
        }
    }

    // Find a pixel format with no antialiasing, if not needed or not supported
    if (AntialiasingLevel == 0)
    {
        // Setup a pixel format descriptor from the rendering settings
        PIXELFORMATDESCRIPTOR PixelDescriptor;
        ZeroMemory(&PixelDescriptor, sizeof(PIXELFORMATDESCRIPTOR));
        PixelDescriptor.nSize        = sizeof(PIXELFORMATDESCRIPTOR);
        PixelDescriptor.nVersion     = 1;
        PixelDescriptor.iLayerType   = PFD_MAIN_PLANE;
        PixelDescriptor.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
        PixelDescriptor.iPixelType   = PFD_TYPE_RGBA;
        PixelDescriptor.cColorBits   = static_cast<BYTE>(Mode.BitsPerPixel);
        PixelDescriptor.cDepthBits   = 32; // It's only a hint
        PixelDescriptor.cStencilBits = 8;  // It's only a hint

        // Get the pixel format that best matches our requirements
        PixelFormat = ChoosePixelFormat(myDeviceContext, &PixelDescriptor);
        if (PixelFormat == 0)
        {
            std::cerr << "Failed to find a suitable pixel format for device context -- cannot create OpenGL context" << std::endl;
            return;
        }
    }

    // Extract the depth and stencil bits from the chosen format
    PIXELFORMATDESCRIPTOR ActualDescriptor;
    ActualDescriptor.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
    ActualDescriptor.nVersion = 1;
    DescribePixelFormat(myDeviceContext, PixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &ActualDescriptor);
    myDepthBits   = ActualDescriptor.cDepthBits;
    myStencilBits = ActualDescriptor.cStencilBits;

    // Set the chosen pixel format
    if (!SetPixelFormat(myDeviceContext, PixelFormat, &ActualDescriptor))
    {
        std::cerr << "Failed to set pixel format for device context -- cannot create OpenGL context" << std::endl;
        return;
    }

    // Create the OpenGL context from the device context
    myGLContext = wglCreateContext(myDeviceContext);
    if (myGLContext == NULL)
    {
        std::cerr << "Failed to create an OpenGL context for this window" << std::endl;
        return;
    }

    // Share display lists with other contexts
    if (wglGetCurrentContext())
        wglShareLists(wglGetCurrentContext(), myGLContext);

    // Set our context as the current OpenGL context for rendering
    SetCurrent();

    // Enable multisampling
    if (AntialiasingLevel > 0)
        glEnable(GL_MULTISAMPLE_ARB);
}


////////////////////////////////////////////////////////////
/// Free all the graphical resources attached to the window
////////////////////////////////////////////////////////////
void WindowImplWin32::Cleanup()
{
    // Restore the previous video mode (in case we were running in fullscreen)
    if (ourFullscreenWindow == this)
    {
        ChangeDisplaySettings(NULL, 0);
        ourFullscreenWindow = NULL;
    }

    // Unhide the mouse cursor (in case it was hidden)
    ShowMouseCursor(true);

    // Destroy the OpenGL context
    if (myGLContext)
    {
        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
        if (this != ourDummyWindow)
        {
            if (ourDummyWindow)
                ourDummyWindow->SetCurrent();
        }
        else
        {
            ourDummyWindow = NULL;
            wglMakeCurrent(NULL, NULL);
        }

        // Destroy the context
        wglDeleteContext(myGLContext);
        myGLContext = NULL;
    }

    // Release the device context
    if (myDeviceContext)
    {
        ReleaseDC(myHandle, myDeviceContext);
        myDeviceContext = NULL;
    }
}


////////////////////////////////////////////////////////////
/// Process a Win32 event
////////////////////////////////////////////////////////////
void WindowImplWin32::ProcessEvent(UINT Message, WPARAM WParam, LPARAM LParam)
{
    // Don't process any message until window is created
    if (myHandle == NULL)
        return;

    switch (Message)
    {
        // Destroy event
        case WM_DESTROY :
        {
            // Here we must cleanup resources !
            Cleanup();
            break;
        }

        // Set cursor event
        case WM_SETCURSOR :
        {
            // The mouse has moved, if the cursor is in our window we must refresh the cursor
            if (LOWORD(LParam) == HTCLIENT)
                SetCursor(myCursor);

            break;
        }

        // Close event
        case WM_CLOSE :
        {
            Event Evt;
            Evt.Type = Event::Close;
            SendEvent(Evt);
            break;
        }

        // Resize event
        case WM_SIZE :
        {
            // Update window size
            RECT Rect;
            GetClientRect(myHandle, &Rect);
            myWidth  = Rect.right - Rect.left;
            myHeight = Rect.bottom - Rect.top;

            Event Evt;
            Evt.Type        = Event::Resize;
            Evt.Size.Width  = myWidth;
            Evt.Size.Height = myHeight;
            SendEvent(Evt);
            break;
        }

        // Gain focus event
        case WM_SETFOCUS :
        {
            Event Evt;
            Evt.Type = Event::GainedFocus;
            SendEvent(Evt);
            break;
        }

        // Lost focus event
        case WM_KILLFOCUS :
        {
            Event Evt;
            Evt.Type = Event::LostFocus;
            SendEvent(Evt);
            break;
        }

        // Text event
        case WM_CHAR :
        {
            Event Evt;
            Evt.Type = Event::TextEntered;
            Evt.Text.Unicode = static_cast<Uint16>(WParam);
            SendEvent(Evt);
            break;
        }

        // Keydown event
        case WM_KEYDOWN :
        case WM_SYSKEYDOWN :
        {
            Event Evt;
            Evt.Type        = Event::KeyPressed;
            Evt.Key.Code    = VirtualKeyCodeToSF(WParam);
            Evt.Key.Alt     = GetAsyncKeyState(VK_RMENU)    || GetAsyncKeyState(VK_LMENU);
            Evt.Key.Control = GetAsyncKeyState(VK_RCONTROL) || GetAsyncKeyState(VK_LCONTROL);
            Evt.Key.Shift   = GetAsyncKeyState(VK_RSHIFT)   || GetAsyncKeyState(VK_LSHIFT);
            SendEvent(Evt);
            break;
        }

        // Keyup event
        case WM_KEYUP :
        case WM_SYSKEYUP :
        {
            Event Evt;
            Evt.Type        = Event::KeyReleased;
            Evt.Key.Code    = VirtualKeyCodeToSF(WParam);
            Evt.Key.Alt     = GetAsyncKeyState(VK_RMENU)    || GetAsyncKeyState(VK_LMENU);
            Evt.Key.Control = GetAsyncKeyState(VK_RCONTROL) || GetAsyncKeyState(VK_LCONTROL);
            Evt.Key.Shift   = GetAsyncKeyState(VK_RSHIFT)   || GetAsyncKeyState(VK_LSHIFT);
            SendEvent(Evt);
            break;
        }

        // Mouse wheel event
        case WM_MOUSEWHEEL :
        {
            Event Evt;
            Evt.Type = Event::MouseWheelMoved;
            Evt.MouseWheel.Delta = static_cast<Int16>(HIWORD(WParam)) / WHEEL_DELTA;
            SendEvent(Evt);
            break;
        }

        // Mouse left button down event
        case WM_LBUTTONDOWN :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonPressed;
            Evt.Mouse.Buttons = Mouse::Left;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse left button up event
        case WM_LBUTTONUP :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonReleased;
            Evt.Mouse.Buttons = Mouse::Left;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse right button down event
        case WM_RBUTTONDOWN :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonPressed;
            Evt.Mouse.Buttons = Mouse::Right;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse right button up event
        case WM_RBUTTONUP :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonReleased;
            Evt.Mouse.Buttons = Mouse::Right;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse wheel button down event
        case WM_MBUTTONDOWN :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonPressed;
            Evt.Mouse.Buttons = Mouse::Middle;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse wheel button up event
        case WM_MBUTTONUP :
        {
            Event Evt;
            Evt.Type          = Event::MouseButtonReleased;
            Evt.Mouse.Buttons = Mouse::Middle;
            Evt.Mouse.X       = LOWORD(LParam);
            Evt.Mouse.Y       = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Mouse move event
        case WM_MOUSEMOVE :
        {
            Event Evt;
            Evt.Type = Event::MouseMove;
            Evt.Mouse.Buttons = 0;
            if (GetAsyncKeyState(VK_LBUTTON)) Evt.Mouse.Buttons |= Mouse::Left;
            if (GetAsyncKeyState(VK_RBUTTON)) Evt.Mouse.Buttons |= Mouse::Right;
            if (GetAsyncKeyState(VK_MBUTTON)) Evt.Mouse.Buttons |= Mouse::Middle;
            Evt.Mouse.X = LOWORD(LParam);
            Evt.Mouse.Y = HIWORD(LParam);
            SendEvent(Evt);
            break;
        }

        // Joystick XY move event
        case MM_JOY1MOVE :
        case MM_JOY2MOVE :
        {
            Event Evt;
            Evt.Type                = Event::JoystickMove;
            Evt.Joystick.JoystickId = Message == MM_JOY1MOVE ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Evt.Joystick.JoystickId];
            Evt.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Evt.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Evt.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            SendEvent(Evt);
            break;
        }

        // Joystick Z move event
        case MM_JOY1ZMOVE :
        case MM_JOY2ZMOVE :
        {
            Event Evt;
            Evt.Type                = Event::JoystickMove;
            Evt.Joystick.JoystickId = Message == MM_JOY1ZMOVE ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Evt.Joystick.JoystickId];
            Evt.Joystick.X = (JoyState.wXpos - Caps.CenterX) * 200 / Caps.RangeX;
            Evt.Joystick.Y = (JoyState.wYpos - Caps.CenterY) * 200 / Caps.RangeY;
            Evt.Joystick.Z = (LOWORD(LParam) - Caps.CenterZ) * 200 / Caps.RangeZ;

            SendEvent(Evt);
            break;
        }

        // Joystick button down events
        case MM_JOY1BUTTONDOWN :
        case MM_JOY2BUTTONDOWN :
        {
            Event Evt;
            Evt.Type                = Event::JoystickButtonPressed;
            Evt.Joystick.JoystickId = Message == MM_JOY1BUTTONDOWN ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Evt.Joystick.JoystickId];
            Evt.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Evt.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Evt.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            if      ((WParam & JOY_BUTTON1) && (WParam & JOY_BUTTON1CHG)) Evt.Joystick.Button = 0;
            else if ((WParam & JOY_BUTTON2) && (WParam & JOY_BUTTON2CHG)) Evt.Joystick.Button = 1;
            else if ((WParam & JOY_BUTTON3) && (WParam & JOY_BUTTON3CHG)) Evt.Joystick.Button = 2;
            else if ((WParam & JOY_BUTTON4) && (WParam & JOY_BUTTON4CHG)) Evt.Joystick.Button = 3;

            SendEvent(Evt);
            break;
        }

        // Joystick button up events
        case MM_JOY1BUTTONUP :
        case MM_JOY2BUTTONUP :
        {
            Event Evt;
            Evt.Type                = Event::JoystickButtonReleased;
            Evt.Joystick.JoystickId = Message == MM_JOY1BUTTONUP ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Evt.Joystick.JoystickId];
            Evt.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Evt.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Evt.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            if      (!(WParam & JOY_BUTTON1) && (WParam & JOY_BUTTON1CHG)) Evt.Joystick.Button = 0;
            else if (!(WParam & JOY_BUTTON2) && (WParam & JOY_BUTTON2CHG)) Evt.Joystick.Button = 1;
            else if (!(WParam & JOY_BUTTON3) && (WParam & JOY_BUTTON3CHG)) Evt.Joystick.Button = 2;
            else if (!(WParam & JOY_BUTTON4) && (WParam & JOY_BUTTON4CHG)) Evt.Joystick.Button = 3;

            SendEvent(Evt);
            break;
        }
    }
}


////////////////////////////////////////////////////////////
/// Setup the joysticks
////////////////////////////////////////////////////////////
void WindowImplWin32::SetupJoysticks()
{
    // Get the number of connected joysticks
    MMRESULT Error;
    JOYINFOEX JoyInfo;
    unsigned int Count = 0;
    while ((Count < 2) && (Error = joyGetPosEx(JOYSTICKID1 + Count, &JoyInfo)) != JOYERR_PARMS)
    {
        // Check if the current joystick is connected
        if (Error == JOYERR_NOERROR)
        {
            // Start capturing joystick
            joySetCapture(myHandle, JOYSTICKID1 + Count, 10, TRUE);

            // Store its caps
            JOYCAPS WinCaps;
            joyGetDevCaps(JOYSTICKID1 + Count, &WinCaps, sizeof(WinCaps));
            myJoystickCaps[Count].CenterX = (WinCaps.wXmax + WinCaps.wXmin) / 2;
            myJoystickCaps[Count].CenterY = (WinCaps.wYmax + WinCaps.wYmin) / 2;
            myJoystickCaps[Count].CenterZ = (WinCaps.wZmax + WinCaps.wZmin) / 2;
            myJoystickCaps[Count].RangeX  = WinCaps.wXmax - WinCaps.wXmin;
            myJoystickCaps[Count].RangeY  = WinCaps.wYmax - WinCaps.wYmin;
            myJoystickCaps[Count].RangeZ  = WinCaps.wZmax - WinCaps.wZmin;
        }

        // Go to the next joystick
        ++Count;
    }
}


////////////////////////////////////////////////////////////
/// Convert a Win32 virtual key code to a SFML key code
////////////////////////////////////////////////////////////
Key::Code WindowImplWin32::VirtualKeyCodeToSF(WPARAM VirtualKey)
{
    switch (VirtualKey)
    {
        case VK_ESCAPE :   return Key::Escape;
        case VK_SPACE :    return Key::Space;
        case VK_RETURN :   return Key::Return;
        case VK_BACK :     return Key::Back;
        case VK_TAB :      return Key::Tab;
        case VK_PRIOR :    return Key::PageUp;
        case VK_NEXT :     return Key::PageDown;
        case VK_END :      return Key::End;
        case VK_HOME :     return Key::Home;
        case VK_INSERT :   return Key::Insert;
        case VK_DELETE :   return Key::Delete;
        case VK_ADD :      return Key::Add;
        case VK_SUBTRACT : return Key::Subtract;
        case VK_MULTIPLY : return Key::Multiply;
        case VK_DIVIDE :   return Key::Divide;
        case VK_PAUSE :    return Key::Pause;
        case VK_F1 :       return Key::F1;
        case VK_F2 :       return Key::F2;
        case VK_F3 :       return Key::F3;
        case VK_F4 :       return Key::F4;
        case VK_F5 :       return Key::F5;
        case VK_F6 :       return Key::F6;
        case VK_F7 :       return Key::F7;
        case VK_F8 :       return Key::F8;
        case VK_F9 :       return Key::F9;
        case VK_F10 :      return Key::F10;
        case VK_F11 :      return Key::F11;
        case VK_F12 :      return Key::F12;
        case VK_F13 :      return Key::F13;
        case VK_F14 :      return Key::F14;
        case VK_F15 :      return Key::F15;
        case VK_LEFT :     return Key::Left;
        case VK_RIGHT :    return Key::Right;
        case VK_UP :       return Key::Up;
        case VK_DOWN :     return Key::Down;
        case VK_NUMPAD0 :  return Key::Numpad0;
        case VK_NUMPAD1 :  return Key::Numpad1;
        case VK_NUMPAD2 :  return Key::Numpad2;
        case VK_NUMPAD3 :  return Key::Numpad3;
        case VK_NUMPAD4 :  return Key::Numpad4;
        case VK_NUMPAD5 :  return Key::Numpad5;
        case VK_NUMPAD6 :  return Key::Numpad6;
        case VK_NUMPAD7 :  return Key::Numpad7;
        case VK_NUMPAD8 :  return Key::Numpad8;
        case VK_NUMPAD9 :  return Key::Numpad9;
        case 'A' :         return Key::A;
        case 'Z' :         return Key::Z;
        case 'E' :         return Key::E;
        case 'R' :         return Key::R;
        case 'T' :         return Key::T;
        case 'Y' :         return Key::Y;
        case 'U' :         return Key::U;
        case 'I' :         return Key::I;
        case 'O' :         return Key::O;
        case 'P' :         return Key::P;
        case 'Q' :         return Key::Q;
        case 'S' :         return Key::S;
        case 'D' :         return Key::D;
        case 'F' :         return Key::F;
        case 'G' :         return Key::G;
        case 'H' :         return Key::H;
        case 'J' :         return Key::J;
        case 'K' :         return Key::K;
        case 'L' :         return Key::L;
        case 'M' :         return Key::M;
        case 'W' :         return Key::W;
        case 'X' :         return Key::X;
        case 'C' :         return Key::C;
        case 'V' :         return Key::V;
        case 'B' :         return Key::B;
        case 'N' :         return Key::N;
        case '0' :         return Key::Num0;
        case '1' :         return Key::Num1;
        case '2' :         return Key::Num2;
        case '3' :         return Key::Num3;
        case '4' :         return Key::Num4;
        case '5' :         return Key::Num5;
        case '6' :         return Key::Num6;
        case '7' :         return Key::Num7;
        case '8' :         return Key::Num8;
        case '9' :         return Key::Num9;
    }

    return Key::Code(0);
}


////////////////////////////////////////////////////////////
/// Function called whenever one of our windows receives a message
////////////////////////////////////////////////////////////
LRESULT CALLBACK WindowImplWin32::GlobalOnEvent(HWND Handle, UINT Message, WPARAM WParam, LPARAM LParam)
{
    // Associate handle and Window instance when the creation message is received
    if (Message == WM_CREATE)
    {
        // Get sfWindowImplWin32 instance (it was passed as the last argument of CreateWindow)
        long This = reinterpret_cast<long>(reinterpret_cast<CREATESTRUCT*>(LParam)->lpCreateParams);

        // Set as the "user data" parameter of the window
        SetWindowLongPtr(Handle, GWLP_USERDATA, This);
    }

    // Get the WindowImpl instance corresponding to the window handle
    WindowImplWin32* Window = reinterpret_cast<WindowImplWin32*>(GetWindowLongPtr(Handle, GWLP_USERDATA));

    // Forward the event to the appropriate function
    if (Window)
    {
        Window->ProcessEvent(Message, WParam, LParam);

        if (Window->myCallback)
            return CallWindowProc(reinterpret_cast<WNDPROC>(Window->myCallback), Handle, Message, WParam, LParam);
    }

    return DefWindowProc(Handle, Message, WParam, LParam);
}

} // namespace priv

} // namespace sf
